#pragma once

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// Copyright (c) 2013 by Nicolas Bonneel (nbonneel@seas.harvard.edu)
/// Computes the curvature of a curve defined in a space of affine transformations
/// whose scaling part is a 1x1 or 2x2 symmetric positive definite matrix in the Wasserstein space
/// and the translational part is Euclidean. This handles several matrices per time step that can 
/// be of up to two different types (typically, a set of 2x3 chrominance affine transforms 
//  and a set of 1x2 luminance affine transforms, per video frame)
///
/// If you use this code, please cite the following paper:
/// Example-Based Video Color Grading
/// Nicolas Bonneel, Kalyan Sunkavalli, Sylvain Paris, Hanspeter Pfister
/// ACM Transactions on Graphics (Proceedings of ACM SIGGRAPH 2013)
/// 
///
/// Redistribution and use in source and binary forms, with or without 
/// modification, are permitted provided that the following conditions are met:
///
/// - Redistributions of source code must retain the above copyright notice,
///   this list of conditions and the following disclaimer.
/// - Redistributions in binary form must reproduce the above copyright notice, 
///   this list of conditions and the following disclaimer in the documentation
///   and/or other materials provided with the distribution.
/// - The name of the author must not be used to endorse or promote products 
///   derived from this software without specific prior written permission.
///
/// THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED 
/// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
/// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO 
/// EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
/// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
/// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
/// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
/// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
/// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
/// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "math_util.h"

// in the following, I use covariance matrix and symmetric positive definite matrix interchangeably


// metric g_V(K, L) for 1x1 covariance matrices  : see supplemental material for the equation
inline double G_1d(double v0, double k0, double l0) {
	return k0*l0/(4*v0);
}
// directional derivative of metric g_V(K, L) in direction D for covariance 1x1 matrices  : see supplemental material for the equation
inline double DG_1d(double d0, double v0, double k0, double l0) {
	return -k0*l0/(4*v0*v0);
}

// metric g_V(K, L) for covariance 2x2 matrices  : see supplemental material for the equation
inline double G_2d(double v0, double v1, double v2, double k0, double k1, double k2, double l0, double l1, double l2) {
return (((-v1*v1+(v0+v2)*v0)*l2-2*v1*v0*l1+v1*v1*l0)*k2+(-2*v2*v1*l0+4*v2*v0*l1-2*l2*v0*v1)*k1+(-2*v2*v1*l1+v1*v1*l2+(-v1*v1+v2*v2+v0*v2)*l0)*k0) / ((v0+v2)*(v0*v2-v1*v1)*4.);
}

// directional derivative of metric g_V(K, L) in direction D for covariance 2x2 matrices  : see supplemental material for the equation
inline double DG_2d(double d0, double d1, double d2, double v0, double v1, double v2, double k0, double k1, double k2, double l0, double l1, double l2) {
	double v0p3 = v0*v0*v0;
	double v1p3 = v1*v1*v1;
	double v1p4 = v1p3*v1;
	double v2p3 = v2*v2*v2;
	double v2p4 = v2p3*v2;
	double denom1 = (v0+v2)*(v0*v2-v1*v1);

	return ((k2-k0)*(l0-l2)*(d0+d2)*v1p4 + 2*(k1*(l2-l0)+(k2-k0)*l1)*(d0*v2-d2*v0)*v1p3-((2*(k2*l1+k1*l2)*d1+(k2*l0+4*k1*l1-(k2-k0)*l2)*d2+k2*l2*d0)*v0*v0+2*((k2*l0-(k2-k0)*l2)*d2+(k1*(l0+l2)+(k0+k2)*l1)*d1 
		+ ((k2-k0)*l0+k0*l2)*d0)*v2*v0+(k0*l0*d2+(k0*l2+4*k1*l1+(k2 - k0)*l0)*d0+2*(k1*l0+k0*l1)*d1)*v2*v2)*v1*v1+(2*((k2*l1+k1*l2)*(2*d2+d0)+((k0+k2)*l2+k2*l0+4*k1*l1)*d1)*v2*v0*v0+2*((k1*l0+k0*l1)*(d2+2*d0)
		+ (k0*l2+4*k1*l1+(k0+k2)*l0)*d1)*v2*v2*v0+2*(k2*l1+k1*l2)*d2*v0p3+2*(k0*l0*d1+(k1*l0+k0*l1)*d0)*v2p3)*v1+k2*l2*v0p3*(2*d1*v1-d2*v0)-2*((k2*l1+k1*l2)*d1+k2*l2*d2)*v2*v0p3-2*((k1*l0+k0*l1)*d1+k0*l0*d0)*v2p3*v0
		- (2*(k1*(l0+l2)+(k0+k2)*l1)*d1+(4*k1*l1+k2*l2)*d2+(4*k1*l1+k0*l0)*d0)*v2*v2*v0*v0-k0*v2p4*l0*d0)/(denom1*denom1*4.);
}

template<int DIM, typename T> double G(const AffineCovarianceMatrix<DIM, T> &V, const AffineCovarianceMatrix<DIM, T> &K, const AffineCovarianceMatrix<DIM, T> &L) {
	switch(DIM){
	case 1:	
		return G_1d(V[0], K[0], L[0]);
		break;
	case 2:		
		return G_2d(V[0], V[1], V[3], K[0], K[1], K[3], L[0], L[1], L[3]);
		break;
	default:
		std::cout<<"currently, only supports 1x1 and 2x2 covariance matrices"<<std::endl;
		return -1;
	}
}

template<int DIM, typename T> double DG(const AffineCovarianceMatrix<DIM, T> &D, const AffineCovarianceMatrix<DIM, T> &V, const AffineCovarianceMatrix<DIM, T> &K, const AffineCovarianceMatrix<DIM, T> &L) {

	switch(DIM) {
	case 1:	
		return DG_1d(D[0], V[0], K[0], L[0]);
		break;
	case 2:		
		return DG_2d(D[0], D[1], D[3], V[0], V[1], V[3], K[0], K[1], K[3], L[0], L[1], L[3]);
		break;
	default:
		std::cout<<"currently, only supports 1x1 and 2x2 covariance matrices"<<std::endl;
		return -1;
	}
}


// takes a serie of nbtransformsperframe affine transformations per frame and returns the square norm of the covariance derivatives of the tangent vector field along the curve
// the affine transformations should consist in a symmetric positive definite n*n matrix and a n*1 translation vectors.
template<typename T, int DIM_SCALE_TM, int DIM_TRANS_TM> 
class GetNablaXX {

public:
	GetNablaXX(const T* alltransforms, int nbframes, int nbtransformsperframe) : transforms(alltransforms) {
		nb_frames = nbframes;
		nb_transforms_per_frame = nbtransformsperframe;

		dTransforms.resize(nb_frames*nb_transforms_per_frame);
		d2Transforms.resize(nb_frames*nb_transforms_per_frame);

		// centered finite differences for each frame		
		for (int i=1; i<nb_frames-1; i++) {		
			for (int j=0; j<nb_transforms_per_frame; j++) {
				dTransforms[i*nb_transforms_per_frame+j] = 0.5*(transforms[(i+1)*nb_transforms_per_frame+j] - transforms[(i-1)*nb_transforms_per_frame+j]); // first derivative
				d2Transforms[i*nb_transforms_per_frame+j] = transforms[(i+1)*nb_transforms_per_frame+j] - 2.0*transforms[i*nb_transforms_per_frame+j] + transforms[(i-1)*nb_transforms_per_frame+j]; // central second derivatives
			}
		}

		// boundaries
		for (int j=0; j<nb_transforms_per_frame; j++) {
			dTransforms[0*nb_transforms_per_frame+j] = transforms[(1)*nb_transforms_per_frame+j] - transforms[(0)*nb_transforms_per_frame+j];
			dTransforms[(nb_frames-1)*nb_transforms_per_frame+j] = transforms[(nb_frames-1)*nb_transforms_per_frame+j] - transforms[(nb_frames-2)*nb_transforms_per_frame+j];
		}
		for (int j=0; j<nb_transforms_per_frame; j++) {
			d2Transforms[0*nb_transforms_per_frame+j] = dTransforms[(1)*nb_transforms_per_frame+j] - dTransforms[(0)*nb_transforms_per_frame+j];
			d2Transforms[(nb_frames-1)*nb_transforms_per_frame+j] = dTransforms[(nb_frames-1)*nb_transforms_per_frame+j] - dTransforms[(nb_frames-2)*nb_transforms_per_frame+j];
		}
		

	};



	double operator()(int frame) {

		double result = 0;

		for (int n=0; n<nb_transforms_per_frame; n++) {
			
			const T& V = transforms[frame*nb_transforms_per_frame+n];     // current transformation
			const T& D = dTransforms[frame*nb_transforms_per_frame+n];    // current first derivative
			const T& E = d2Transforms[frame*nb_transforms_per_frame+n];   // current second derivative
			
			
			// Get the basis vectors (the standard basis for symmetric matrices)
			T basis_scale[DIM_SCALE_TM];
			for (int i=0; i<DIM_SCALE_TM; i++) {
				basis_scale[i] = T::get_basis_scale(i);
			}

			// Computes pairwise "dot products" between all these basis vectors (this "dot product" actually uses the Wasserstein metric)
			double g_scale[DIM_SCALE_TM*DIM_SCALE_TM];
			for (int i=0; i<DIM_SCALE_TM; i++) {
				for (int j=0; j<DIM_SCALE_TM; j++) {
					g_scale[i*DIM_SCALE_TM+j] = G(V, basis_scale[i], basis_scale[j]);
				}
			}
			

			// inverse of the resulting matrix
			double inv_g_scale[DIM_SCALE_TM*DIM_SCALE_TM];
			mat_inv<double>(g_scale, inv_g_scale, DIM_SCALE_TM);


			// derivative of the above matrix in the direction of all the basis vectors
			double DerivG_scale[DIM_SCALE_TM*DIM_SCALE_TM*DIM_SCALE_TM];			
			for (int j=0; j<DIM_SCALE_TM; j++) {
				for (int k=0; k<DIM_SCALE_TM; k++) {
					for (int l=0; l<DIM_SCALE_TM; l++) {
						DerivG_scale[j*DIM_SCALE_TM*DIM_SCALE_TM + k*DIM_SCALE_TM + l] = DG(basis_scale[j], V, basis_scale[k], basis_scale[l]);       // DerivG_scale[j][k][l] = d[g(basis_k, basis_l)] / d[basis_j]
					}
				}
			}
			

			// Computes the Christoffel symbols
			double Gamma_scale[DIM_SCALE_TM*DIM_SCALE_TM*DIM_SCALE_TM];
			memset(Gamma_scale, 0, DIM_SCALE_TM*DIM_SCALE_TM*DIM_SCALE_TM*sizeof(Gamma_scale[0]));
			for (int l=0; l<DIM_SCALE_TM; l++) {
				for (int j=0; j<DIM_SCALE_TM; j++) {
					for (int k=0; k<DIM_SCALE_TM; k++) {
						for (int r=0; r<DIM_SCALE_TM; r++) {
							Gamma_scale[l*DIM_SCALE_TM*DIM_SCALE_TM + j*DIM_SCALE_TM + k] += 0.5*inv_g_scale[l*DIM_SCALE_TM + r] * (DerivG_scale[k*DIM_SCALE_TM*DIM_SCALE_TM + r*DIM_SCALE_TM + j] + DerivG_scale[j*DIM_SCALE_TM*DIM_SCALE_TM + r*DIM_SCALE_TM + k] - DerivG_scale[r*DIM_SCALE_TM*DIM_SCALE_TM + j*DIM_SCALE_TM + k]);
						}
					}
				}
			}


			T covariant_deriv = T::Zero();

			// starts with the second derivative as the covariant derivative
			covariant_deriv += E;  
			
			// and adds the non-Euclidean terms			
			for (int k=0; k<DIM_SCALE_TM; k++) {
				for (int i=0; i<DIM_SCALE_TM; i++) {
					for (int l=0; l<DIM_SCALE_TM; l++) {
						double val =  (Gamma_scale[k*DIM_SCALE_TM*DIM_SCALE_TM + i*DIM_SCALE_TM + l]*D[T::base_to_index(i)])*D[T::base_to_index(l)];  // Wiki: http://en.wikipedia.org/wiki/Geodesic				
						int row = T::base_to_index(k)/T::dim;
						int col = T::base_to_index(k)%T::dim;
						covariant_deriv[row*T::dim+col] +=val;  // just corresponds to multiplying val by the n'th basis vector (i.e., basis matrix)
						if (row!=col)
							covariant_deriv[col*T::dim+row] +=val;
					}
				}
			}

			result+=G(V, covariant_deriv, covariant_deriv);  // we are interested in the (squared) norm of the covariant derivative. This is calculated in the Wasserstein space ...
			result+=dot_trans(E, E); // ... apart from the translation component of the affine transformation, which is handled in a Euclidean way

		}

		return result;
	}


	

private:
	const T* transforms;	
	int nb_frames;
	int nb_transforms_per_frame;

	std::vector<T> dTransforms;
	std::vector<T> d2Transforms;

};




// takes a serie of N=nbtransformslumperframe matrices per frame for the chrominance, and M=nbtransformschromperframe matrices per frame for the luminance, and operator()(frame) results in the curvature at frame "frame"
// the matrices should be of type AffineCovarianceMatrix, which is basically a double* in which the n*n first elements are the covariance matrix (symmetric, so row major or column major is the same) and the next n elements are the translational components
// for an n*n matrix, DIM_SCALE_TM1=n*(n+1)/2  and DIM_TRANS_TM1=n  
// the code only handles 1x1 and 2x2 matrices for now (+ translations)
template<typename T1, int DIM_SCALE_TM1, int DIM_TRANS_TM1, typename T2, int DIM_SCALE_TM2, int DIM_TRANS_TM2> 
class Curvature {

public:
	Curvature(const T1* transformslum, const T2* transformschrom, int nbframes, int nbtransformslumperframe, int nbtransformschromperframe) : 
		  GetNabla1(transformslum, nbframes, nbtransformslumperframe),
		  GetNabla2(transformschrom, nbframes, nbtransformschromperframe)
	  {
	};

	double operator()(int frame) {

		double Nabla1(GetNabla1(frame)); 
		double Nabla2(GetNabla2(frame));
		double result = Nabla1 + Nabla2;
		
		if (_finite(result) && result>0)
			return sqrt(result);
		else 
			return 0;		
	}

	GetNablaXX<T1, DIM_SCALE_TM1, DIM_TRANS_TM1> GetNabla1;
	GetNablaXX<T2, DIM_SCALE_TM2, DIM_TRANS_TM2> GetNabla2;
	
};